@@ -1,46 +1,11 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the QtQml module of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:LGPL$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 3 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL3 included in the
-** packaging of this file. Please review the following information to
-** ensure the GNU Lesser General Public License version 3 requirements
-** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 2.0 or (at your option) the GNU General
-** Public license version 3 or any later version approved by the KDE Free
-** Qt Foundation. The licenses are as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-2.0.html and
-** https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2016 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
 #include "qtqmlworkerscriptglobal_p.h"
 #include "qquickworkerscript_p.h"
 #include <private/qqmlengine_p.h>
 #include <private/qqmlexpression_p.h>
+#include <private/qjsvalue_p.h>
 
 #include <QtCore/qcoreevent.h>
 #include <QtCore/qcoreapplication.h>
@@ -124,16 +89,21 @@ private:
     QQmlError m_error;
 };
 
-struct WorkerScript : public QV4::ExecutionEngine {
-    WorkerScript(int id, QQuickWorkerScriptEnginePrivate *parent);
+struct WorkerScript : public QV4::ExecutionEngine::Deletable
+{
+    WorkerScript(QV4::ExecutionEngine *);
+    ~WorkerScript() = default;
 
     QQuickWorkerScriptEnginePrivate *p = nullptr;
     QUrl source;
     QQuickWorkerScript *owner = nullptr;
+#if QT_CONFIG(qml_network)
     QScopedPointer<QNetworkAccessManager> scriptLocalNAM;
-    int id = -1;
+#endif
 };
 
+V4_DEFINE_EXTENSION(WorkerScript, workerScriptExtension);
+
 class QQuickWorkerScriptEnginePrivate : public QObject
 {
     Q_OBJECT
@@ -149,12 +119,16 @@ public:
     QMutex m_lock;
     QWaitCondition m_wait;
 
-    QHash<int, WorkerScript *> workers;
-    QV4::ReturnedValue getWorker(WorkerScript *);
+    // ExecutionEngines are owned by the worker script and created and deleted
+    // in the worker thread. QQuickWorkerScript instances, however, belong to
+    // the main thread. They are only inserted as place holders when creating
+    // the worker script.
+    QHash<int, QBiPointer<QV4::ExecutionEngine, QQuickWorkerScript>> workers;
 
     int m_nextId;
 
     static QV4::ReturnedValue method_sendMessage(const QV4::FunctionObject *, const QV4::Value *thisObject, const QV4::Value *argv, int argc);
+    QV4::ExecutionEngine *workerEngine(int id);
 
 signals:
     void stopThread();
@@ -177,13 +151,14 @@ QV4::ReturnedValue QQuickWorkerScriptEnginePrivate::method_sendMessage(const QV4
                                                                        const QV4::Value *, const QV4::Value *argv, int argc)
 {
     QV4::Scope scope(b);
-    WorkerScript *script = static_cast<WorkerScript *>(scope.engine);
+    const WorkerScript *script = workerScriptExtension(scope.engine);
+    Q_ASSERT(script);
 
     QV4::ScopedValue v(scope, argc > 0 ? argv[0] : QV4::Value::undefinedValue());
     QByteArray data = QV4::Serialize::serialize(v, scope.engine);
 
     QMutexLocker locker(&script->p->m_lock);
-    if (script && script->owner)
+    if (script->owner)
         QCoreApplication::postEvent(script->owner, new WorkerDataEvent(0, data));
 
     return QV4::Encode::undefined();
@@ -205,9 +180,10 @@ bool QQuickWorkerScriptEnginePrivate::event(QEvent *event)
     } else if (event->type() == (QEvent::Type)WorkerRemoveEvent::WorkerRemove) {
         QMutexLocker locker(&m_lock);
         WorkerRemoveEvent *workerEvent = static_cast<WorkerRemoveEvent *>(event);
-        QHash<int, WorkerScript *>::iterator itr = workers.find(workerEvent->workerId());
-        if (itr != workers.end()) {
-            delete itr.value();
+        auto itr = workers.constFind(workerEvent->workerId());
+        if (itr != workers.cend()) {
+            if (itr->isT1())
+                delete itr->asT1();
             workers.erase(itr);
         }
         return true;
@@ -216,30 +192,48 @@ bool QQuickWorkerScriptEnginePrivate::event(QEvent *event)
     }
 }
 
+QV4::ExecutionEngine *QQuickWorkerScriptEnginePrivate::workerEngine(int id)
+{
+    const auto it = workers.find(id);
+    if (it == workers.end())
+        return nullptr;
+    if (it->isT1())
+        return it->asT1();
+
+    QQuickWorkerScript *owner = it->asT2();
+    auto *engine = new QV4::ExecutionEngine;
+    WorkerScript *script = workerScriptExtension(engine);
+    script->owner = owner;
+    script->p = this;
+    *it = engine;
+    return engine;
+}
+
 void QQuickWorkerScriptEnginePrivate::processMessage(int id, const QByteArray &data)
 {
-    WorkerScript *script = workers.value(id);
-    if (!script)
+    QV4::ExecutionEngine *engine = workerEngine(id);
+    if (!engine)
         return;
 
-    QV4::Scope scope(script);
-    QV4::ScopedString v(scope);
-    QV4::ScopedObject worker(scope, script->globalObject->get((v = script->newString(QStringLiteral("WorkerScript")))));
+    QV4::Scope scope(engine);
+    QV4::ScopedString v(scope, engine->newString(QStringLiteral("WorkerScript")));
+    QV4::ScopedObject worker(scope, engine->globalObject->get(v));
     QV4::ScopedFunctionObject onmessage(scope);
     if (worker)
-        onmessage = worker->get((v = script->newString(QStringLiteral("onMessage"))));
+        onmessage = worker->get((v = engine->newString(QStringLiteral("onMessage"))));
 
     if (!onmessage)
         return;
 
-    QV4::ScopedValue value(scope, QV4::Serialize::deserialize(data, script));
+    QV4::ScopedValue value(scope, QV4::Serialize::deserialize(data, engine));
 
-    QV4::JSCallData jsCallData(scope, 1);
-    *jsCallData->thisObject = script->global();
-    jsCallData->args[0] = value;
+    QV4::JSCallArguments jsCallData(scope, 1);
+    *jsCallData.thisObject = engine->global();
+    jsCallData.args[0] = value;
     onmessage->call(jsCallData);
     if (scope.hasException()) {
         QQmlError error = scope.engine->catchExceptionAsQmlError();
+        WorkerScript *script = workerScriptExtension(engine);
         reportScriptException(script, error);
     }
 }
@@ -251,39 +245,41 @@ void QQuickWorkerScriptEnginePrivate::processLoad(int id, const QUrl &url)
 
     QString fileName = QQmlFile::urlToLocalFileOrQrc(url);
 
-    WorkerScript *script = workers.value(id);
-    if (!script)
+    QV4::ExecutionEngine *engine = workerEngine(id);
+    if (!engine)
         return;
 
+    WorkerScript *script = workerScriptExtension(engine);
     script->source = url;
 
     if (fileName.endsWith(QLatin1String(".mjs"))) {
-        auto moduleUnit = script->loadModule(url);
-        if (moduleUnit) {
-            if (moduleUnit->instantiate(script))
-                moduleUnit->evaluate();
+        auto module = engine->loadModule(url);
+        if (module.compiled) {
+            if (module.compiled->instantiate())
+                module.compiled->evaluate();
+        } else if (module.native) {
+            // Nothing to do. There is no global code in a native module.
         } else {
-            script->throwError(QStringLiteral("Could not load module file"));
+            engine->throwError(QStringLiteral("Could not load module file"));
         }
     } else {
         QString error;
-        QV4::Scope scope(script);
+        QV4::Scope scope(engine);
         QScopedPointer<QV4::Script> program;
-        program.reset(QV4::Script::createFromFileOrCache(script, /*qmlContext*/nullptr, fileName, url, &error));
+        program.reset(QV4::Script::createFromFileOrCache(
+                          engine, /*qmlContext*/nullptr, fileName, url, &error));
         if (program.isNull()) {
             if (!error.isEmpty())
                 qWarning().nospace() << error;
             return;
         }
 
-        if (!script->hasException)
+        if (!engine->hasException)
             program->run();
     }
 
-    if (script->hasException) {
-        QQmlError error = script->catchExceptionAsQmlError();
-        reportScriptException(script, error);
-    }
+    if (engine->hasException)
+        reportScriptException(script, engine->catchExceptionAsQmlError());
 }
 
 void QQuickWorkerScriptEnginePrivate::reportScriptException(WorkerScript *script,
@@ -378,20 +374,25 @@ QQuickWorkerScriptEngine::~QQuickWorkerScriptEngine()
     delete d;
 }
 
-WorkerScript::WorkerScript(int id, QQuickWorkerScriptEnginePrivate *parent)
-    : p(parent)
-    , id(id)
-{
-    initQmlGlobalObject();
 
-    QV4::Scope scope(this);
-    QV4::ScopedObject api(scope, scope.engine->newObject());
-    QV4::ScopedString name(scope, newString(QStringLiteral("sendMessage")));
-    QV4::ScopedValue sendMessage(scope, QV4::FunctionObject::createBuiltinFunction(this, name, QQuickWorkerScriptEnginePrivate::method_sendMessage, 1));
-    api->put(QV4::ScopedString(scope, scope.engine->newString(QStringLiteral("sendMessage"))), sendMessage);
-    globalObject->put(QV4::ScopedString(scope, scope.engine->newString(QStringLiteral("WorkerScript"))), api);
-    networkAccessManager = [](QV4::ExecutionEngine *engine){
-        auto *workerScript = static_cast<WorkerScript *>(engine);
+WorkerScript::WorkerScript(QV4::ExecutionEngine *engine)
+{
+    engine->initQmlGlobalObject();
+
+    QV4::Scope scope(engine);
+    QV4::ScopedObject api(scope, engine->newObject());
+    QV4::ScopedString sendMessageName(scope, engine->newString(QStringLiteral("sendMessage")));
+    QV4::ScopedFunctionObject sendMessage(
+                scope, QV4::FunctionObject::createBuiltinFunction(
+                    engine, sendMessageName,
+                    QQuickWorkerScriptEnginePrivate::method_sendMessage, 1));
+    api->put(sendMessageName, sendMessage);
+    QV4::ScopedString workerScriptName(scope, engine->newString(QStringLiteral("WorkerScript")));
+    engine->globalObject->put(workerScriptName, api);
+
+#if QT_CONFIG(qml_network)
+    engine->networkAccessManager = [](QV4::ExecutionEngine *engine) {
+        WorkerScript *workerScript = workerScriptExtension(engine);
         if (workerScript->scriptLocalNAM)
             return workerScript->scriptLocalNAM.get();
         if (auto *namFactory = workerScript->p->qmlengine->networkAccessManagerFactory())
@@ -400,27 +401,31 @@ WorkerScript::WorkerScript(int id, QQuickWorkerScriptEnginePrivate *parent)
             workerScript->scriptLocalNAM.reset(new QNetworkAccessManager(workerScript->p));
         return workerScript->scriptLocalNAM.get();
     };
+#endif // qml_network
 }
 
 int QQuickWorkerScriptEngine::registerWorkerScript(QQuickWorkerScript *owner)
 {
-    WorkerScript *script = new WorkerScript(d->m_nextId++, d);
-
-    script->owner = owner;
+    const int id = d->m_nextId++;
 
     d->m_lock.lock();
-    d->workers.insert(script->id, script);
+    d->workers.insert(id, owner);
     d->m_lock.unlock();
 
-    return script->id;
+    return id;
 }
 
 void QQuickWorkerScriptEngine::removeWorkerScript(int id)
 {
-    if (WorkerScript *script = d->workers.value(id)) {
-        script->owner = nullptr;
-        QCoreApplication::postEvent(d, new WorkerRemoveEvent(id));
+    const auto it = d->workers.constFind(id);
+    if (it == d->workers.cend())
+        return;
+
+    if (it->isT1()) {
+        QV4::ExecutionEngine *engine = it->asT1();
+        workerScriptExtension(engine)->owner = nullptr;
     }
+    QCoreApplication::postEvent(d, new WorkerRemoveEvent(id));
 }
 
 void QQuickWorkerScriptEngine::executeUrl(int id, const QUrl &url)
@@ -436,21 +441,23 @@ void QQuickWorkerScriptEngine::sendMessage(int id, const QByteArray &data)
 void QQuickWorkerScriptEngine::run()
 {
     d->m_lock.lock();
-
     d->m_wait.wakeAll();
-
     d->m_lock.unlock();
 
     exec();
 
-    qDeleteAll(d->workers);
+    for (auto it = d->workers.begin(), end = d->workers.end(); it != end; ++it) {
+        if (it->isT1())
+            delete it->asT1();
+    }
+
     d->workers.clear();
 }
 
 
 /*!
     \qmltype WorkerScript
-    \instantiates QQuickWorkerScript
+    \nativetype QQuickWorkerScript
     \ingroup qtquick-threading
     \inqmlmodule QtQml.WorkerScript
     \brief Enables the use of threads in a Qt Quick application.
@@ -501,9 +508,6 @@ void QQuickWorkerScriptEngine::run()
 
     Worker scripts that are plain JavaScript sources can not use \l {qtqml-javascript-imports.html}{.import} syntax.
     Scripts that are ECMAScript modules can freely use import and export statements.
-
-    \sa {Qt Quick Examples - Threading},
-        {Threaded ListModel Example}
 */
 QQuickWorkerScript::QQuickWorkerScript(QObject *parent)
 : QObject(parent), m_engine(nullptr), m_scriptId(-1), m_componentComplete(true)
@@ -537,12 +541,25 @@ void QQuickWorkerScript::setSource(const QUrl &source)
 
     m_source = source;
 
-    if (engine())
-        m_engine->executeUrl(m_scriptId, m_source);
+    if (engine()) {
+        const QQmlContext *context = qmlContext(this);
+        m_engine->executeUrl(m_scriptId, context ? context->resolvedUrl(m_source) : m_source);
+    }
 
     emit sourceChanged();
 }
 
+/*!
+    \qmlproperty bool WorkerScript::ready
+
+    This holds whether the WorkerScript has been initialized and is ready
+    for receiving messages via \tt WorkerScript.sendMessage().
+*/
+bool QQuickWorkerScript::ready() const
+{
+    return m_engine != nullptr;
+}
+
 /*!
     \qmlmethod WorkerScript::sendMessage(jsobject message)
 
@@ -563,7 +580,7 @@ void QQuickWorkerScript::setSource(const QUrl &source)
     of ListModel objects, any modifications by the other thread to an object
     passed in \c message will not be reflected in the original object.
 */
-void QQuickWorkerScript::sendMessage(QQmlV4Function *args)
+void QQuickWorkerScript::sendMessage(QQmlV4FunctionPtr args)
 {
     if (!engine()) {
         qWarning("QQuickWorkerScript: Attempt to send message before WorkerScript establishment");
@@ -587,7 +604,8 @@ QQuickWorkerScriptEngine *QQuickWorkerScript::engine()
 {
     if (m_engine) return m_engine;
     if (m_componentComplete) {
-        QQmlEngine *engine = qmlEngine(this);
+        const QQmlContext *context = qmlContext(this);
+        QQmlEngine *engine = context ? context->engine() : nullptr;
         if (!engine) {
             qWarning("QQuickWorkerScript: engine() called without qmlEngine() set");
             return nullptr;
@@ -601,7 +619,9 @@ QQuickWorkerScriptEngine *QQuickWorkerScript::engine()
         m_scriptId = m_engine->registerWorkerScript(this);
 
         if (m_source.isValid())
-            m_engine->executeUrl(m_scriptId, m_source);
+            m_engine->executeUrl(m_scriptId, context->resolvedUrl(m_source));
+
+        emit readyChanged();
 
         return m_engine;
     }
@@ -619,8 +639,6 @@ void QQuickWorkerScript::componentComplete()
 
     This signal is emitted when a message \a msg is received from a worker
     script in another thread through a call to sendMessage().
-
-    The corresponding handler is \c onMessage.
 */
 
 bool QQuickWorkerScript::event(QEvent *event)
@@ -629,7 +647,8 @@ bool QQuickWorkerScript::event(QEvent *event)
         if (QQmlEngine *engine = qmlEngine(this)) {
             QV4::ExecutionEngine *v4 = engine->handle();
             WorkerDataEvent *workerEvent = static_cast<WorkerDataEvent *>(event);
-            emit message(QJSValue(v4, QV4::Serialize::deserialize(workerEvent->data(), v4)));
+            emit message(QJSValuePrivate::fromReturnedValue(
+                             QV4::Serialize::deserialize(workerEvent->data(), v4)));
         }
         return true;
     } else if (event->type() == (QEvent::Type)WorkerErrorEvent::WorkerError) {
