@@ -1,30 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 Canonical Limited and/or its subsidiary(-ies).
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the test suite of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:GPL-EXCEPT$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 as published by the Free Software
-** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2016 Canonical Limited and/or its subsidiary(-ies).
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
 #include <QtTest/QtTest>
 #include <QtQml/qqmlengine.h>
@@ -38,13 +13,19 @@
 #include <QtQml/private/qqmlengine_p.h>
 #include <QtQml/private/qqmltypedata_p.h>
 #include <QtQml/private/qqmltypeloader_p.h>
-#include "../../shared/testhttpserver.h"
-#include "../../shared/util.h"
+#include <QtQml/private/qqmlirbuilder_p.h>
+#include <QtQml/private/qqmlirloader_p.h>
+#include <QtQuickTestUtils/private/testhttpserver_p.h>
+#include <QtQuickTestUtils/private/qmlutils_p.h>
+#include <QQmlComponent>
 
 class tst_QQMLTypeLoader : public QQmlDataTest
 {
     Q_OBJECT
 
+public:
+    tst_QQMLTypeLoader();
+
 private slots:
     void testLoadComplete();
     void loadComponentSynchronously();
@@ -57,27 +38,44 @@ private slots:
     void redirect();
     void qmlSingletonWithinModule();
     void multiSingletonModule();
+    void multiSingletonModuleNoWarning();
     void implicitComponentModule();
+    void customDiskCachePath();
     void qrcRootPathUrl();
     void implicitImport();
     void compositeSingletonCycle();
+    void declarativeCppType();
+    void circularDependency();
+    void declarativeCppAndQmlDir();
+    void signalHandlersAreCompatible();
+    void loadTypeOnShutdown();
+
+private:
+    void checkSingleton(const QString & dataDirectory);
 };
 
+tst_QQMLTypeLoader::tst_QQMLTypeLoader()
+    : QQmlDataTest(QT_QMLTEST_DATADIR)
+{
+}
+
 void tst_QQMLTypeLoader::testLoadComplete()
 {
-    QQuickView *window = new QQuickView();
+#ifdef Q_OS_ANDROID
+    QSKIP("Loading dynamic plugins does not work on Android");
+#endif
+    std::unique_ptr<QQuickView> window = std::make_unique<QQuickView>();
     window->engine()->addImportPath(QT_TESTCASE_BUILDDIR);
     qDebug() << window->engine()->importPathList();
     window->setGeometry(0,0,240,320);
     window->setSource(testFileUrl("test_load_complete.qml"));
     window->show();
-    QVERIFY(QTest::qWaitForWindowExposed(window));
+    QVERIFY(QTest::qWaitForWindowExposed(window.get()));
 
     QObject *rootObject = window->rootObject();
     QTRY_VERIFY(rootObject != nullptr);
     QTRY_COMPARE(rootObject->property("created").toInt(), 2);
     QTRY_COMPARE(rootObject->property("loaded").toInt(), 2);
-    delete window;
 }
 
 void tst_QQMLTypeLoader::loadComponentSynchronously()
@@ -94,11 +92,17 @@ void tst_QQMLTypeLoader::trimCache()
 {
     QQmlEngine engine;
     QQmlTypeLoader &loader = QQmlEnginePrivate::get(&engine)->typeLoader;
+    QVector<QQmlTypeData *> releaseLater;
+    QVector<QV4::CompiledData::CompilationUnit *> releaseCompilationUnitLater;
     for (int i = 0; i < 256; ++i) {
         QUrl url = testFileUrl("trim_cache.qml");
         url.setQuery(QString::number(i));
 
         QQmlTypeData *data = loader.getType(url).take();
+
+        // Backup source code should be dropped right after loading, even without cache trimming.
+        QVERIFY(!data->backupSourceCode().isValid());
+
         // Run an event loop to receive the callback that release()es.
         QTRY_COMPARE(data->count(), 2);
 
@@ -106,8 +110,10 @@ void tst_QQMLTypeLoader::trimCache()
         // QQmlTypeData or its compiledData() should prevent the trimming.
         if (i % 10 == 0) {
             // keep ref on data, don't add ref on data->compiledData()
+            releaseLater.append(data);
         } else if (i % 5 == 0) {
             data->compilationUnit()->addref();
+            releaseCompilationUnitLater.append(data->compilationUnit());
             data->release();
         } else {
             data->release();
@@ -123,6 +129,12 @@ void tst_QQMLTypeLoader::trimCache()
             QVERIFY(!loader.isTypeLoaded(url));
         // The cache is free to keep the others.
     }
+
+    for (auto *data : std::as_const(releaseCompilationUnitLater))
+        data->release();
+
+    for (auto *data : std::as_const(releaseLater))
+        data->release();
 }
 
 void tst_QQMLTypeLoader::trimCache2()
@@ -158,7 +170,7 @@ void tst_QQMLTypeLoader::trimCache3()
     QCOMPARE(loader.isTypeLoaded(testFileUrl("ComponentWithIncubator.qml")), false);
 }
 
-static void checkSingleton(const QString &dataDirectory)
+void tst_QQMLTypeLoader::checkSingleton(const QString &dataDirectory)
 {
     QQmlEngine engine;
     engine.addImportPath(dataDirectory);
@@ -167,8 +179,8 @@ static void checkSingleton(const QString &dataDirectory)
                       "import QtQuick 2.6\n"
                       "import \"..\"\n"
                       "Item { property int t: ValueSource.something }",
-                      QUrl::fromLocalFile(dataDirectory + "/abc/Xyz.qml"));
-    QCOMPARE(component.status(), QQmlComponent::Ready);
+                      testFileUrl("abc/Xyz.qml"));
+    QVERIFY2(component.status() == QQmlComponent::Ready, qPrintable(component.errorString()));
     QScopedPointer<QObject> o(component.create());
     QVERIFY(o.data());
     QCOMPARE(o->property("t").toInt(), 10);
@@ -252,7 +264,7 @@ public:
             return;
         m_buffer.clear();
         setError(ContentNotFoundError, "content not found");
-        emit error(ContentNotFoundError);
+        emit errorOccurred(ContentNotFoundError);
         setFinished(true);
         emit finished();
     }
@@ -264,8 +276,8 @@ public:
 
     qint64 readData(char *data, qint64 maxlen) override
     {
-        if (m_buffer.length() < maxlen)
-            maxlen = m_buffer.length();
+        if (m_buffer.size() < maxlen)
+            maxlen = m_buffer.size();
         std::memcpy(data, m_buffer.data(), maxlen);
         m_buffer.remove(0, maxlen);
         return maxlen;
@@ -306,7 +318,8 @@ public:
 
         NetworkReply *reply = new NetworkReply;
         QString filename = QQmlFile::urlToLocalFileOrQrc(url);
-        QTimer::singleShot(10, reply, [this, reply, filename]() {
+        QTimer::singleShot(QRandomGenerator::global()->bounded(20), reply,
+                           [this, reply, filename]() {
             if (filename.isEmpty()) {
                 reply->fail();
             } else {
@@ -334,10 +347,14 @@ public:
             if (line.isEmpty())
                 continue;
             QList<QByteArray> segments = line.split(' ');
+            if (segments.startsWith("optional")) {
+                result.append("optional ");
+                segments.removeFirst();
+            }
             if (segments.startsWith("plugin")) {
-                if (segments.length() == 2) {
+                if (segments.size() == 2) {
                     segments.append(path);
-                } else if (segments.length() == 3) {
+                } else if (segments.size() == 3) {
                     if (!segments[2].startsWith('/'))
                         segments[2] = path + segments[2];
                 } else {
@@ -371,6 +388,19 @@ public:
     }
 };
 
+class ManualRedirectNetworkAccessManagerFactory : public QQmlNetworkAccessManagerFactory
+{
+public:
+    QStringList loadedFiles;
+
+    QNetworkAccessManager *create(QObject *parent) override
+    {
+        NetworkAccessManager *manager = new NetworkAccessManager(parent);
+        manager->setRedirectPolicy(QNetworkRequest::ManualRedirectPolicy);
+        return manager;
+    }
+};
+
 class UrlInterceptor : public QQmlAbstractUrlInterceptor
 {
 public:
@@ -390,6 +420,9 @@ public:
 
 void tst_QQMLTypeLoader::intercept()
 {
+#ifdef Q_OS_ANDROID
+    QSKIP("Loading dynamic plugins does not work on Android");
+#endif
     qmlClearTypeRegistrations();
 
     QQmlEngine engine;
@@ -399,7 +432,7 @@ void tst_QQMLTypeLoader::intercept()
     UrlInterceptor interceptor;
     NetworkAccessManagerFactory factory;
 
-    engine.setUrlInterceptor(&interceptor);
+    engine.addUrlInterceptor(&interceptor);
     engine.setNetworkAccessManagerFactory(&factory);
 
     QQmlComponent component(&engine, testFileUrl("test_intercept.qml"));
@@ -414,7 +447,7 @@ void tst_QQMLTypeLoader::intercept()
     QTRY_COMPARE(o->property("created").toInt(), 2);
     QTRY_COMPARE(o->property("loaded").toInt(), 2);
 
-    QVERIFY(factory.loadedFiles.length() >= 6);
+    QVERIFY(factory.loadedFiles.size() >= 6);
     QVERIFY(factory.loadedFiles.contains(dataDirectory() + "/test_intercept.qml"));
     QVERIFY(factory.loadedFiles.contains(dataDirectory() + "/Intercept.qml"));
     QVERIFY(factory.loadedFiles.contains(dataDirectory() + "/Fast/qmldir"));
@@ -430,7 +463,9 @@ void tst_QQMLTypeLoader::redirect()
     QVERIFY(server.serveDirectory(dataDirectory()));
     server.addRedirect("Base.qml", server.urlString("/redirected/Redirected.qml"));
 
+    ManualRedirectNetworkAccessManagerFactory factory;
     QQmlEngine engine;
+    engine.setNetworkAccessManagerFactory(&factory);
     QQmlComponent component(&engine);
     component.loadUrl(server.urlString("/Load.qml"), QQmlComponent::Asynchronous);
     QTRY_VERIFY2(component.isReady(), qPrintable(component.errorString()));
@@ -479,6 +514,9 @@ static void checkCleanCacheLoad(const QString &testCase)
 
 void tst_QQMLTypeLoader::multiSingletonModule()
 {
+#ifdef Q_OS_ANDROID
+    QSKIP("Android seems to have problems with QProcess");
+#endif
     qmlClearTypeRegistrations();
     QQmlEngine engine;
     engine.addImportPath(testFile("imports"));
@@ -497,8 +535,23 @@ void tst_QQMLTypeLoader::multiSingletonModule()
     checkCleanCacheLoad(QLatin1String("multiSingletonModule"));
 }
 
+void tst_QQMLTypeLoader::multiSingletonModuleNoWarning()
+{
+    // Should not warn about a "cyclic" dependency between the singletons
+    QTest::failOnWarning(QRegularExpression(".*"));
+
+    QQmlEngine engine;
+    QQmlComponent component(&engine, testFileUrl("imports/multisingletonmodule/a.qml"));
+    QVERIFY2(component.isReady(), qPrintable(component.errorString()));
+    QScopedPointer<QObject> o(component.create());
+    QVERIFY(!o.isNull());
+}
+
 void tst_QQMLTypeLoader::implicitComponentModule()
 {
+#ifdef Q_OS_ANDROID
+    QSKIP("Android seems to have problems with QProcess");
+#endif
     QQmlEngine engine;
     QQmlComponent component(&engine, testFileUrl("implicitcomponent.qml"));
     QCOMPARE(component.status(), QQmlComponent::Ready);
@@ -508,6 +561,39 @@ void tst_QQMLTypeLoader::implicitComponentModule()
     checkCleanCacheLoad(QLatin1String("implicitComponentModule"));
 }
 
+void tst_QQMLTypeLoader::customDiskCachePath()
+{
+#ifdef Q_OS_ANDROID
+    QSKIP("Android seems to have problems with QProcess");
+#endif
+
+#if QT_CONFIG(process)
+    const char *skipKey = "QT_TST_QQMLTYPELOADER_SKIP_MISMATCH";
+    if (qEnvironmentVariableIsSet(skipKey)) {
+        QQmlEngine engine;
+        QQmlComponent component(&engine, testFileUrl("Base.qml"));
+        QCOMPARE(component.status(), QQmlComponent::Ready);
+        QScopedPointer<QObject> obj(component.create());
+        QVERIFY(!obj.isNull());
+        return;
+    }
+
+    QTemporaryDir dir;
+    QProcess child;
+    child.setProgram(QCoreApplication::applicationFilePath());
+    child.setArguments(QStringList(QLatin1String("customDiskCachePath")));
+    QProcessEnvironment env = QProcessEnvironment::systemEnvironment();
+    env.insert(QLatin1String(skipKey), QLatin1String("1"));
+    env.insert(QLatin1String("QML_DISK_CACHE_PATH"), dir.path());
+    child.setProcessEnvironment(env);
+    child.start();
+    QVERIFY(child.waitForFinished());
+    QCOMPARE(child.exitCode(), 0);
+    QDir cacheDir(dir.path());
+    QVERIFY(!cacheDir.isEmpty());
+#endif
+}
+
 void tst_QQMLTypeLoader::qrcRootPathUrl()
 {
     QQmlEngine engine;
@@ -519,10 +605,25 @@ void tst_QQMLTypeLoader::implicitImport()
 {
     QQmlEngine engine;
     engine.addImportPath(testFile("imports"));
-    QQmlComponent component(&engine, testFileUrl("implicitimporttest.qml"));
-    QVERIFY2(component.isReady(), qPrintable(component.errorString()));
-    QScopedPointer<QObject> obj(component.create());
-    QVERIFY(!obj.isNull());
+    {
+        QQmlComponent component(&engine, testFileUrl("implicitimporttest.qml"));
+        QVERIFY2(component.isReady(), qPrintable(component.errorString()));
+        QScopedPointer<QObject> obj(component.create());
+        QVERIFY(!obj.isNull());
+    }
+    {
+        QQmlComponent component(&engine, testFileUrl("implicitautoimporttest.qml"));
+        QVERIFY2(component.isReady(), qPrintable(component.errorString()));
+        QScopedPointer<QObject> obj(component.create());
+        QVERIFY(!obj.isNull());
+    }
+    {
+        QQmlComponent component(&engine, testFileUrl("implicitversionedimporttest.qml"));
+        QVERIFY2(component.isReady(), qPrintable(component.errorString()));
+        QScopedPointer<QObject> obj(component.create());
+        QVERIFY(!obj.isNull());
+    }
+
 }
 
 void tst_QQMLTypeLoader::compositeSingletonCycle()
@@ -542,6 +643,143 @@ void tst_QQMLTypeLoader::compositeSingletonCycle()
     QCOMPARE(qvariant_cast<QColor>(object->property("color")), QColorConstants::Black);
 }
 
+void tst_QQMLTypeLoader::declarativeCppType()
+{
+    QQmlEngine engine;
+    QQmlComponent component(&engine, testFileUrl("declarativeCppType.qml"));
+    QCOMPARE(component.status(), QQmlComponent::Ready);
+    QScopedPointer<QObject> obj(component.create());
+    QVERIFY(!obj.isNull());
+}
+
+void tst_QQMLTypeLoader::circularDependency()
+{
+    QQmlEngine engine;
+    QTest::ignoreMessage(QtWarningMsg, QRegularExpression("Cyclic dependency detected between (.*) and (.*)"));
+    QQmlComponent component(&engine, testFileUrl("CircularDependency.qml"));
+    QCOMPARE(component.status(), QQmlComponent::Null);
+}
+
+void tst_QQMLTypeLoader::declarativeCppAndQmlDir()
+{
+    QQmlEngine engine;
+    engine.addImportPath("qrc:/");
+    QQmlComponent component(&engine, testFileUrl("cppAndQmlDir.qml"));
+    QVERIFY2(!component.isError(), qPrintable(component.errorString()));
+    QScopedPointer<QObject> root(component.create());
+    QCOMPARE(root->objectName(), "Singleton");
+}
+
+static void getCompilationUnitAndRuntimeInfo(QQmlRefPointer<QV4::ExecutableCompilationUnit> &unit,
+                                             QList<int> &runtimeFunctionIndices, const QUrl &url,
+                                             QQmlEngine *engine)
+{
+    QQmlTypeLoader &loader = QQmlEnginePrivate::get(engine)->typeLoader;
+    auto typeData = loader.getType(url);
+    QVERIFY(typeData);
+    QVERIFY(!typeData->backupSourceCode().isValid());
+
+    if (typeData->isError()) {
+        const auto errors = typeData->errors();
+        for (const QQmlError &e : errors)
+            qDebug().noquote() << e.toString();
+        QVERIFY(!typeData->isError()); // this returns
+    }
+
+    unit = engine->handle()->executableCompilationUnit(typeData->compilationUnit());
+    QVERIFY(unit);
+
+    // the QmlIR::Document is deleted once loader.getType() is complete, so
+    // restore it
+    QmlIR::Document restoredIrDocument(false);
+    QQmlIRLoader irLoader(unit->unitData(), &restoredIrDocument);
+    irLoader.load();
+    QCOMPARE(restoredIrDocument.objects.size(), 1);
+
+    const QmlIR::Object *irRoot = restoredIrDocument.objects.at(0);
+    runtimeFunctionIndices = QList<int>(irRoot->runtimeFunctionIndices.begin(),
+                                        irRoot->runtimeFunctionIndices.end());
+}
+
+void tst_QQMLTypeLoader::signalHandlersAreCompatible()
+{
+    QQmlEngine engine;
+
+    QQmlRefPointer<QV4::ExecutableCompilationUnit> unitFromCachegen;
+    QList<int> runtimeFunctionIndicesFromCachegen;
+    getCompilationUnitAndRuntimeInfo(unitFromCachegen, runtimeFunctionIndicesFromCachegen,
+                                     // use qmlcachegen version
+                                     QUrl("qrc:/data/compilercompatibility/signalHandlers.qml"),
+                                     &engine);
+    if (QTest::currentTestFailed())
+        return;
+
+    QQmlRefPointer<QV4::ExecutableCompilationUnit> unitFromTypeCompiler;
+    QList<int> runtimeFunctionIndicesFromTypeCompiler;
+    getCompilationUnitAndRuntimeInfo(unitFromTypeCompiler, runtimeFunctionIndicesFromTypeCompiler,
+                                     // use qqmltypecompiler version
+                                     testFileUrl("compilercompatibility/signalHandlers.qml"),
+                                     &engine);
+    if (QTest::currentTestFailed())
+        return;
+
+    // this is a "bare minimum" test, but if this succeeds, we could test other
+    // things elsewhere
+    QCOMPARE(runtimeFunctionIndicesFromCachegen, runtimeFunctionIndicesFromTypeCompiler);
+    QCOMPARE(unitFromCachegen->runtimeFunctions.size(),
+             unitFromTypeCompiler->runtimeFunctions.size());
+    // make sure that units really come from different places (the machinery
+    // could in theory be smart enough to figure the qmlcachegen cached
+    // version), fairly questionable check but better than nothing
+#ifdef Q_OS_ANDROID
+    QSKIP("qrc and file system is the same thing on Android");
+#endif
+    QVERIFY(unitFromCachegen->url() != unitFromTypeCompiler->url());
+}
+
+void tst_QQMLTypeLoader::loadTypeOnShutdown()
+{
+    bool dead1 = false;
+    bool dead2 = false;
+
+    {
+        QQmlEngine engine;
+        auto good = new QQmlComponent(
+                &engine, testFileUrl("doesExist.qml"),
+                QQmlComponent::CompilationMode::Asynchronous, &engine);
+        QObject::connect(
+                good, &QQmlComponent::statusChanged, &engine,
+                [&](QQmlComponent::Status) {
+
+            // Must not call this if the engine is already dead.
+            QVERIFY(engine.rootContext());
+
+        });
+
+        QObject::connect(good, &QQmlComponent::destroyed, good, [&]() { dead1 = true; });
+        QVERIFY(good->isLoading());
+
+        auto bad = new QQmlComponent(
+                &engine, testFileUrl("doesNotExist.qml"),
+                QQmlComponent::CompilationMode::Asynchronous, &engine);
+        QObject::connect(
+                bad, &QQmlComponent::statusChanged, &engine,
+                [&](QQmlComponent::Status) {
+
+            // Must not call this if the engine is already dead.
+            // Must also not leak memory from the events the error produces.
+            QVERIFY(engine.rootContext());
+
+        });
+
+        QObject::connect(bad, &QQmlComponent::destroyed, bad, [&]() { dead2 = true; });
+        QVERIFY(bad->isLoading());
+    }
+
+    QVERIFY(dead1);
+    QVERIFY(dead2);
+}
+
 QTEST_MAIN(tst_QQMLTypeLoader)
 
 #include "tst_qqmltypeloader.moc"
